@@ -55,6 +55,8 @@
 #include <osmocom/sgsn/gprs_ranap.h>
 #include <osmocom/sgsn/gprs_gmm_fsm.h>
 #include <osmocom/sgsn/gprs_mm_state_gb_fsm.h>
+#include <osmocom/sgsn/gtp_mme.h>
+#include <osmocom/sgsn/sgsn_rim.h>
 
 #include <gtp.h>
 #include <pdp.h>
@@ -167,6 +169,7 @@ struct sgsn_pdp_ctx *sgsn_create_pdp_ctx(struct sgsn_ggsn_ctx *ggsn,
 	pdp->hisaddr0 =	ggsn->remote_addr;
 	pdp->hisaddr1 = ggsn->remote_addr;
 	//pdp->cch_pdp = 512;	/* Charging Flat Rate */
+	pdp->radio_pri = 0x4;
 
 	/* MS provided APN, subscription was verified by the caller */
 	pdp->selmode = 0xFC | 0x00;
@@ -225,18 +228,11 @@ struct sgsn_pdp_ctx *sgsn_create_pdp_ctx(struct sgsn_ggsn_ctx *ggsn,
 		qos = TLVP_VAL(tp, OSMO_IE_GSM_REQ_QOS);
 	}
 
-	if (qos_len <= 3) {
-		pdp->qos_req.l = qos_len + 1;
-		if (pdp->qos_req.l > sizeof(pdp->qos_req.v))
-			pdp->qos_req.l = sizeof(pdp->qos_req.v);
-		pdp->qos_req.v[0] = 0; /* Allocation/Retention policy */
-		memcpy(&pdp->qos_req.v[1], qos, pdp->qos_req.l - 1);
-	} else {
-		pdp->qos_req.l = qos_len;
-		if (pdp->qos_req.l > sizeof(pdp->qos_req.v))
-			pdp->qos_req.l = sizeof(pdp->qos_req.v);
-		memcpy(pdp->qos_req.v, qos, pdp->qos_req.l);
-	}
+	pdp->qos_req.l = qos_len + 1;
+	if (pdp->qos_req.l > sizeof(pdp->qos_req.v))
+		pdp->qos_req.l = sizeof(pdp->qos_req.v);
+	pdp->qos_req.v[0] = 0; /* Allocation/Retention policy */
+	memcpy(&pdp->qos_req.v[1], qos, pdp->qos_req.l - 1);
 
 	/* charging characteristics if present */
 	if (TLVP_LEN(tp, OSMO_IE_GSM_CHARG_CHAR) >= sizeof(pdp->cch_pdp))
@@ -311,11 +307,16 @@ struct sgsn_pdp_ctx *sgsn_create_pdp_ctx(struct sgsn_ggsn_ctx *ggsn,
 	return pctx;
 }
 
-/* SGSN wants to delete a PDP context */
+/* SGSN wants to delete a PDP context, send first DeleteCtxReq on the GTP side,
+   then upon DeleteCtx ACK it will send DeactPdpAcc to the MS if still
+   connected. */
 int sgsn_delete_pdp_ctx(struct sgsn_pdp_ctx *pctx)
 {
 	LOGPDPCTXP(LOGL_INFO, pctx, "Delete PDP Context\n");
 
+	OSMO_ASSERT(pctx->ggsn);
+	OSMO_ASSERT(pctx->lib);
+
 	/* FIXME: decide if we need teardown or not ! */
 	return gtp_delete_context_req2(pctx->ggsn->gsn, pctx->lib, pctx, 1);
 }
@@ -473,6 +474,70 @@ void sgsn_ggsn_echo_req(struct sgsn_ggsn_ctx *ggc)
 	gtp_echo_req(ggc->gsn, ggc->gtp_version, ggc, &ggc->remote_addr);
 }
 
+int sgsn_mme_ran_info_req(struct sgsn_mme_ctx *mme, const struct bssgp_ran_information_pdu *pdu)
+{
+	char ri_src_str[64], ri_dest_str[64];
+	int ri_len;
+	struct msgb *msg;
+	struct bssgp_normal_hdr *bgph;
+	int rc;
+	uint8_t ri_buf[64];
+	uint8_t *ri_ptr = &ri_buf[0];
+	struct sockaddr_in sk_in = {
+		.sin_family = AF_INET,
+		.sin_port = htons(GTP1C_PORT),
+		.sin_addr = mme->remote_addr,
+	};
+
+	msg = bssgp_encode_rim_pdu(pdu);
+	if (!msg) {
+		LOGMME(mme, DRIM, LOGL_ERROR, "Tx GTP RAN Information Relay: failed to encode pdu\n");
+		return -EINVAL;
+	}
+	bgph = (struct bssgp_normal_hdr *)msgb_bssgph(msg);
+	DEBUGP(DLBSSGP, "Tx GTP RAN Information Relay: RIM-PDU:%s, src=%s, dest=%s\n",
+	       bssgp_pdu_str(bgph->pdu_type),
+	       bssgp_rim_ri_name_buf(ri_src_str, sizeof(ri_src_str), &pdu->routing_info_src),
+	       bssgp_rim_ri_name_buf(ri_dest_str, sizeof(ri_dest_str), &pdu->routing_info_dest));
+
+	if ((ri_len = bssgp_create_rim_ri(ri_ptr, &pdu->routing_info_dest)) < 0) {
+		ri_ptr = NULL;
+		ri_len = 0;
+	}
+
+	rc = gtp_ran_info_relay_req(mme->sgsn->gsn,  &sk_in, msgb_data(msg), msgb_length(msg),
+				    ri_ptr, ri_len, pdu->routing_info_dest.discr);
+	msgb_free(msg);
+	return rc;
+}
+
+/* Confirmation of a PDP Context Update */
+static int update_pdp_conf(struct pdp_t *pdp, void *cbp, int cause)
+{
+	struct sgsn_pdp_ctx *pctx = cbp;
+	int rc;
+
+	LOGPDPCTXP(LOGL_INFO, pctx, "Received Update PDP CTX CONF, cause=%d(%s)\n",
+		cause, get_value_string(gtp_cause_strs, cause));
+
+	/* 3GPP TS 29.060 "7.3.4":
+	 * "If the SGSN receives an Update PDP Context Response with a Cause
+	 * value other than "Request accepted", it shall abort the update of the
+	 * PDP context.""
+	 * "If the SGSN receives an Update PDP Context Response with
+	 * a Cause value "Non-existent", it shall delete the PDP Context."
+	 */
+	if (cause != GTPCAUSE_NON_EXIST)
+		 return 0; /* Nothing to do */
+
+	LOGPDPCTXP(LOGL_INFO, pctx, "PDP CTX we tried to update doesn't exist in "
+		   "the GGSN anymore, deleting it locally.\n");
+
+	rc = gtp_freepdp(pctx->ggsn->gsn, pctx->lib);
+	/* related mmctx is torn down in cb_delete_context called by gtp_freepdp() */
+	return rc;
+}
+
 /* Confirmation of a PDP Context Delete */
 static int delete_pdp_conf(struct pdp_t *pdp, void *cbp, int cause)
 {
@@ -571,6 +636,8 @@ static int cb_conf(int type, int cause, struct pdp_t *pdp, void *cbp)
 		return echo_conf(cbp, cause == EOF);
 	case GTP_CREATE_PDP_REQ:
 		return create_pdp_conf(pdp, cbp, cause);
+	case GTP_UPDATE_PDP_REQ:
+		return update_pdp_conf(pdp, cbp, cause);
 	case GTP_DELETE_PDP_REQ:
 		return delete_pdp_conf(pdp, cbp, cause);
 	default:
@@ -613,6 +680,42 @@ static int cb_extheader_ind(struct sockaddr_in *peer)
 	return 0;
 }
 
+static int cb_gtp_ran_info_relay_ind(struct sockaddr_in *peer, union gtpie_member **ie)
+{
+	char addrbuf[INET_ADDRSTRLEN];
+	struct sgsn_mme_ctx *mme = sgsn_mme_ctx_by_addr(sgsn, &peer->sin_addr);
+	if (!mme) {
+		LOGP(DGTP, LOGL_NOTICE, "Rx GTP RAN Information Relay from unknown MME %s\n",
+		     inet_ntop(AF_INET, &peer->sin_addr, addrbuf, sizeof(addrbuf)));
+		return -ECONNREFUSED;
+	}
+
+	LOGMME(mme, DGTP, LOGL_INFO, "Rx GTP RAN Information Relay\n");
+
+	unsigned int len = 0;
+	struct msgb *msg = msgb_alloc(4096, "gtpcv1_ran_info");
+	struct bssgp_ran_information_pdu pdu;
+
+	if (gtpie_gettlv(ie, GTPIE_RAN_T_CONTAIN, 0, &len, msgb_data(msg), 4096) || len <= 0) {
+		LOGMME(mme, DGTP, LOGL_ERROR, "Rx GTP RAN Information Relay: No Transparent Container IE found!\n");
+		goto ret_error;
+	}
+	msgb_put(msg, len);
+	msgb_bssgph(msg) = msg->data;
+	msgb_nsei(msg) = 0;
+	if (bssgp_parse_rim_pdu(&pdu, msg) < 0) {
+		LOGMME(mme, DGTP, LOGL_ERROR, "Rx GTP RAN Information Relay: Failed parsing Transparent Container IE!\n");
+		goto ret_error;
+	}
+
+	msgb_free(msg);
+	return sgsn_rim_rx_from_gtp(&pdu, mme);
+
+ret_error:
+	msgb_free(msg);
+	return -EINVAL;
+}
+
 /* Called whenever we receive a DATA packet */
 static int cb_data_ind(struct pdp_t *lib, void *packet, unsigned int len)
 {
@@ -660,14 +763,18 @@ static int cb_data_ind(struct pdp_t *lib, void *packet, unsigned int len)
 
 	switch (mm->gmm_fsm->state) {
 	case ST_GMM_REGISTERED_SUSPENDED:
-		/* initiate PS PAGING procedure */
-		gprs_gb_page_ps_ra(mm);
-		/* FIXME: queue the packet we received from GTP */
-		break;
+		LOGMMCTXP(LOGL_INFO, mm, "Dropping DL packet for MS in GMM state %s\n",
+			  osmo_fsm_inst_state_name(mm->gmm_fsm));
+		msgb_free(msg);
+		return -1;
 	case ST_GMM_REGISTERED_NORMAL:
 		OSMO_ASSERT(mm->gb.mm_state_fsm->state != ST_MM_IDLE);
-		if (mm->gb.mm_state_fsm->state == ST_MM_STANDBY)
+		if (mm->gb.mm_state_fsm->state == ST_MM_STANDBY) {
+			LOGMMCTXP(LOGL_INFO, mm, "Paging MS in GMM state %s, MM state %s\n",
+				  osmo_fsm_inst_state_name(mm->gmm_fsm),
+				  osmo_fsm_inst_state_name(mm->gb.mm_state_fsm));
 			gprs_gb_page_ps_ra(mm);
+		}
 
 		/* FIXME: queue the packet we received from GTP */
 		break;
@@ -678,10 +785,10 @@ static int cb_data_ind(struct pdp_t *lib, void *packet, unsigned int len)
 		return -1;
 	}
 
-	rate_ctr_inc(&pdp->ctrg->ctr[PDP_CTR_PKTS_UDATA_OUT]);
-	rate_ctr_add(&pdp->ctrg->ctr[PDP_CTR_BYTES_UDATA_OUT], len);
-	rate_ctr_inc(&mm->ctrg->ctr[GMM_CTR_PKTS_UDATA_OUT]);
-	rate_ctr_add(&mm->ctrg->ctr[GMM_CTR_BYTES_UDATA_OUT], len);
+	rate_ctr_inc(rate_ctr_group_get_ctr(pdp->ctrg, PDP_CTR_PKTS_UDATA_OUT));
+	rate_ctr_add(rate_ctr_group_get_ctr(pdp->ctrg, PDP_CTR_BYTES_UDATA_OUT), len);
+	rate_ctr_inc(rate_ctr_group_get_ctr(mm->ctrg, GMM_CTR_PKTS_UDATA_OUT));
+	rate_ctr_add(rate_ctr_group_get_ctr(mm->ctrg, GMM_CTR_BYTES_UDATA_OUT), len);
 
 	/* It is easier to have a global count */
 	pdp->cdr_bytes_out += len;
@@ -716,10 +823,10 @@ int sgsn_rx_sndcp_ud_ind(struct gprs_ra_id *ra_id, int32_t tlli, uint8_t nsapi,
 		return -EIO;
 	}
 
-	rate_ctr_inc(&pdp->ctrg->ctr[PDP_CTR_PKTS_UDATA_IN]);
-	rate_ctr_add(&pdp->ctrg->ctr[PDP_CTR_BYTES_UDATA_IN], npdu_len);
-	rate_ctr_inc(&mmctx->ctrg->ctr[GMM_CTR_PKTS_UDATA_IN]);
-	rate_ctr_add(&mmctx->ctrg->ctr[GMM_CTR_BYTES_UDATA_IN], npdu_len);
+	rate_ctr_inc(rate_ctr_group_get_ctr(pdp->ctrg, PDP_CTR_PKTS_UDATA_IN));
+	rate_ctr_add(rate_ctr_group_get_ctr(pdp->ctrg, PDP_CTR_BYTES_UDATA_IN), npdu_len);
+	rate_ctr_inc(rate_ctr_group_get_ctr(mmctx->ctrg, GMM_CTR_PKTS_UDATA_IN));
+	rate_ctr_add(rate_ctr_group_get_ctr(mmctx->ctrg, GMM_CTR_BYTES_UDATA_IN), npdu_len);
 
 	/* It is easier to have a global count */
 	pdp->cdr_bytes_in += npdu_len;
@@ -733,7 +840,7 @@ static int sgsn_gtp_fd_cb(struct osmo_fd *fd, unsigned int what)
 	struct sgsn_instance *sgi = fd->data;
 	int rc;
 
-	if (!(what & BSC_FD_READ))
+	if (!(what & OSMO_FD_READ))
 		return 0;
 
 	switch (fd->priv_nr) {
@@ -771,31 +878,19 @@ int sgsn_gtp_init(struct sgsn_instance *sgi)
 	if (gsn->mode != GTP_MODE_SGSN)
 		return -EINVAL;
 
-	sgi->gtp_fd0.fd = gsn->fd0;
-	sgi->gtp_fd0.priv_nr = 0;
-	sgi->gtp_fd0.data = sgi;
-	sgi->gtp_fd0.when = BSC_FD_READ;
-	sgi->gtp_fd0.cb = sgsn_gtp_fd_cb;
+	osmo_fd_setup(&sgi->gtp_fd0, gsn->fd0, OSMO_FD_READ, sgsn_gtp_fd_cb, sgi, 0);
 	rc = osmo_fd_register(&sgi->gtp_fd0);
 	if (rc < 0)
 		return rc;
 
-	sgi->gtp_fd1c.fd = gsn->fd1c;
-	sgi->gtp_fd1c.priv_nr = 1;
-	sgi->gtp_fd1c.data = sgi;
-	sgi->gtp_fd1c.when = BSC_FD_READ;
-	sgi->gtp_fd1c.cb = sgsn_gtp_fd_cb;
+	osmo_fd_setup(&sgi->gtp_fd1c, gsn->fd1c, OSMO_FD_READ, sgsn_gtp_fd_cb, sgi, 1);
 	rc = osmo_fd_register(&sgi->gtp_fd1c);
 	if (rc < 0) {
 		osmo_fd_unregister(&sgi->gtp_fd0);
 		return rc;
 	}
 
-	sgi->gtp_fd1u.fd = gsn->fd1u;
-	sgi->gtp_fd1u.priv_nr = 2;
-	sgi->gtp_fd1u.data = sgi;
-	sgi->gtp_fd1u.when = BSC_FD_READ;
-	sgi->gtp_fd1u.cb = sgsn_gtp_fd_cb;
+	osmo_fd_setup(&sgi->gtp_fd1u, gsn->fd1u, OSMO_FD_READ, sgsn_gtp_fd_cb, sgi, 2);
 	rc = osmo_fd_register(&sgi->gtp_fd1u);
 	if (rc < 0) {
 		osmo_fd_unregister(&sgi->gtp_fd0);
@@ -810,6 +905,7 @@ int sgsn_gtp_init(struct sgsn_instance *sgi)
 	gtp_set_cb_data_ind(gsn, cb_data_ind);
 	gtp_set_cb_unsup_ind(gsn, cb_unsup_ind);
 	gtp_set_cb_extheader_ind(gsn, cb_extheader_ind);
+	gtp_set_cb_ran_info_relay_ind(gsn, cb_gtp_ran_info_relay_ind);
 
 	return 0;
 }
